import math


def log(value: float | int) -> float:
    return math.log10(value)


def exp(value: float | int) -> float:
    return 10.0 ** value


def abs_to_rel(value: int | float, range_: int | float, min_: int | float) -> float:
    return (value - min_) / range_



from __future__ import annotations
import abc
import dataclasses


class Plottable(metaclass=abc.ABCMeta):
    @staticmethod
    def get_marker_chars() -> list[str]:
        pass

    @abc.abstractmethod
    def plot(self) -> None:
        pass


class _CharFieldWarning(str):
    pass


@dataclasses.dataclass(frozen=True)
class TerminalPoint:
    x: int
    y: int

    def __post_init__(self):
        if self.x < 0:
            raise ValueError("Terminal point coordination should be positive: x")
        if self.y < 0:
            raise ValueError("Terminal point coordination should be positive: y")


@dataclasses.dataclass(frozen=True)
class TerminalMarker(TerminalPoint):
    char: str

    def __post_init__(self):
        if len(self.char) != 1:
            raise ValueError("Length of Marker char should be 1")


@dataclasses.dataclass(frozen=True)
class TerminalLabel(TerminalPoint):
    label: str
    allow_left_shift: bool = False

    def __len__(self):
        return len(self.label)


@dataclasses.dataclass(frozen=True)
class TerminalXAxis:
    axis_line: list[TerminalMarker]
    tick_labels: list[TerminalLabel]
    axis_label: TerminalLabel | None = None


@dataclasses.dataclass(frozen=True)
class TerminalYAxis:
    axis_line: list[TerminalMarker]
    tick_labels: list[TerminalLabel]
    axis_label: list[TerminalMarker] | None = None


@dataclasses.dataclass(frozen=True)
class TerminalLegend:
    legend_elements: list[TerminalLabel]


class _CharField:
    def __init__(self, line_num: int, col_num: int):
        self.char_field: list[list[str]] = [[" " for _c in range(col_num)] for _l in range(line_num)]
        self.cfw_list = []

    def write_marker(self, marker: TerminalMarker):
        if self.char_field[marker.y][marker.x] != " ":
            self.cfw_list.append(
                _CharFieldWarning(
                    "Overlapping markers detected. "
                    "If you need more accurate plot, consider changing the terminal size: (x=%d, y=%d)" % (marker.x, marker.y)
                )
            )
        self.char_field[marker.y][marker.x] = marker.char

    def write_label(self, label: TerminalLabel):
        if label.allow_left_shift and (label.x + len(label)) > len(self.char_field[0]):
            # ラベルの左移動が許可されている and ラベルが右にはみ出ている
            shift = label.x + len(label) - len(self.char_field[0])
            label = TerminalLabel(label.x - shift, label.y, label.label)

        y = label.y
        for i in range(len(label)):
            x = label.x + i
            if self.char_field[y][x] != " ":
                self.cfw_list.append(
                    _CharFieldWarning(
                        "Overlapping labels detected. "
                        "If you need more accurate plot, consider changing the terminal size: (x=%d, y=%d)" % (x, y)
                    )
                )
            self.char_field[y][x] = label.label[i]

    def project(self) -> None:
        for w in self.cfw_list:
            print(w)
        for line in reversed(self.char_field):
            print("".join(line))


@dataclasses.dataclass(frozen=True)
class Terminal(Plottable):
    line_num: int
    col_num: int
    plot_markers: list[TerminalMarker]
    x_axis: TerminalXAxis
    y_axis: TerminalYAxis
    legend: TerminalLegend | None

    @staticmethod
    def get_marker_chars() -> list[str]:
        return ["*", "o", "+", "x", "v", "#", "."]

    def plot(self) -> None:
        cf = _CharField(line_num=self.line_num, col_num=self.col_num)

        # x axis
        for marker in self.x_axis.axis_line:
            cf.write_marker(marker)
        for label in self.x_axis.tick_labels:
            cf.write_label(label)
        if self.x_axis.axis_label:
            cf.write_label(self.x_axis.axis_label)

        # y axis
        for marker in self.y_axis.axis_line:
            cf.write_marker(marker)
        for label in self.y_axis.tick_labels:
            cf.write_label(label)
        for marker in self.y_axis.axis_label or []:
            cf.write_marker(marker)

        # legend
        if self.legend:
            for legend in self.legend.legend_elements:
                cf.write_label(legend)

        # marker
        for marker in self.plot_markers:
            cf.write_marker(marker)

        cf.project()



import pytest

import scatterminal.common as common


@pytest.mark.parametrize(
    ("value", "range_", "min_", "expected"),
    [
        (5, 10, 0, 0.5),
        (15, 10, 10, 0.5),
    ]
)
def test_abs_to_rel(value: int | float, range_: int | float, min_: int | float, expected: float):
    assert common.abs_to_rel(value, range_, min_) == expected


@pytest.mark.parametrize(
    ("value",),
    [
        (1,),
        (1.0,),
        (180.5,)
    ]
)
def test_log_exp(value: int | float):
    assert common.exp(common.log(value)) == pytest.approx(float(value))



from typing import Callable

import pytest

import scatterminal.data_layer_model as dlm


@pytest.mark.parametrize(
    ("filter_func", "data_seq", "expected"),
    [
        # positive-pass filter x
        (
            lambda xy: xy[0] > 0,
            dlm.DataSequence(
                x=[1, 2, 3],
                y=[1, 2, 3],
                seq_id=0
            ),
            dlm.DataSequence(
                x=[1, 2, 3],
                y=[1, 2, 3],
                seq_id=0
            )
        ),
        (
            lambda xy: xy[0] > 0,
            dlm.DataSequence(
                x=[1, -2, 3],
                y=[1, 2, 3],
                seq_id=0
            ),
            dlm.DataSequence(
                x=[1, 3],
                y=[1, 3],
                seq_id=0
            )
        ),
        (
            lambda xy: xy[0] > 0,
            dlm.DataSequence(
                x=[1, 2, 3],
                y=[1, -2, 3],
                seq_id=0
            ),
            dlm.DataSequence(
                x=[1, 2, 3],
                y=[1, -2, 3],
                seq_id=0
            )
        ),

        # positive-pass filter y
        (
            lambda xy: xy[1] > 0,
            dlm.DataSequence(
                x=[1, 2, 3],
                y=[1, 2, 3],
                seq_id=0
            ),
            dlm.DataSequence(
                x=[1, 2, 3],
                y=[1, 2, 3],
                seq_id=0
            )
        ),
        (
            lambda xy: xy[1] > 0,
            dlm.DataSequence(
                x=[1, -2, 3],
                y=[1, 2, 3],
                seq_id=0
            ),
            dlm.DataSequence(
                x=[1, -2, 3],
                y=[1, 2, 3],
                seq_id=0
            )
        ),
        (
            lambda xy: xy[1] > 0,
            dlm.DataSequence(
                x=[1, 2, 3],
                y=[1, -2, 3],
                seq_id=0
            ),
            dlm.DataSequence(
                x=[1, 3],
                y=[1, 3],
                seq_id=0
            )
        )
    ]
)
def test_data_sequence_create_filtered(
        filter_func: Callable[[tuple[int | float, int | float]], bool],
        data_seq: dlm.DataSequence,
        expected: dlm.DataSequence
):
    actual = data_seq.create_filtered(filter_func)
    assert actual == expected


@pytest.mark.parametrize(
    ("x", "y", "seq_id", "name", "x_name", "error_type", "expected"),
    [
        ([0], [1, 2], 9, None, None, ValueError, "length of x and y must be equal: (seq_id=9, len(x)=1, len(y)=2)"),
        ([0, "A"], [1, 2], 9, None, None, TypeError, "x must be list[int | float]: seq_id=9"),
        ([0, 1], [1, "A"], 9, None, None, TypeError, "y must be list[int | float]: seq_id=9"),
        ([0, 1], [1, 2], 9, "あ", None, ValueError, "Sequence name must be ascii: seq_id=9"),
        ([0, 1], [1, 2], 9, None, "あ", ValueError, "Sequence x_name must be ascii: seq_id=9")
    ]
)
def test_data_sequence_error(
        x: list[int | str], y: list[int | str],
        seq_id, name: str | None, x_name: str | None,
        error_type: type[Exception], expected: str
):
    with pytest.raises(error_type) as e:
        _ = dlm.DataSequence(x, y, seq_id, name, x_name)
    assert str(e.value) == expected


@pytest.mark.parametrize(
    ("scale", "name", "min_", "max_", "expected"),
    [
        (dlm.DataScaleType.linear, "あ", 0, 1, "Axis name must be ascii: あ"),
        (dlm.DataScaleType.linear, "a", 0, None, "Axis min and Axis max should be defined simultaneously: (min, max)=(0, None)"),
        (dlm.DataScaleType.linear, "a", None, 0, "Axis min and Axis max should be defined simultaneously: (min, max)=(None, 0)"),
        (dlm.DataScaleType.linear, "a", 0, 0, "Axis min is larger than max: (min, max)="),
        (dlm.DataScaleType.log, "a", 0, 10, "Axis min must be positive value on log-scale"),
    ]
)
def test_data_axis_error(scale: dlm.DataScaleType, name: str | None, min_: int | None, max_: int | None, expected: str):
    with pytest.raises(ValueError) as e:
        _ = dlm.DataAxis(scale, name, min_, max_)
    assert expected in str(e.value)


@pytest.mark.parametrize(
    ("data", "expected"),
    [
        (
            [dlm.DataSequence([0, 1], [1, 2], 9)],
            "Non-positive value is detected on log-scale x axis. This data point is not plotted.: seq_id=9"
        ),
        (
            [dlm.DataSequence([1, 1], [0, 2], 9)],
            "Non-positive value is detected on log-scale y axis. This data point is not plotted.: seq_id=9"
        )
    ]
)
def test_data_warning(data: list[dlm.DataSequence], expected: str):
    with pytest.warns(UserWarning) as e:
        _ = dlm.Data(data, dlm.DataAxis(dlm.DataScaleType.log), dlm.DataAxis(dlm.DataScaleType.log), dlm.DataLegendLoc.lower)
    assert str(e.list[0].message) == expected



import argparse
import dataclasses


from scatterminal.canvas_layer_model import Canvas
from scatterminal.csv_parser import read_file, parse
from scatterminal.data_layer_model import DataScaleType, DataAxis, Data, DataLegendLoc, DataSequence, SimpleDataSequence
from scatterminal.terminal_layer_model import Terminal


@dataclasses.dataclass(frozen=True)
class PlotParameter:
    x_scale: DataScaleType
    y_scale: DataScaleType


def plot_csv(
        file_paths: list[str],
        sep: str | None = None,
        x_label: str | None = None,
        y_label: str | None = None,
        x_scale: str = "linear",
        y_scale: str = "linear",
        x_lim: tuple[float, float] | None = None,
        y_lim: tuple[float, float] | None = None,
        legend_loc: str = "lower"
) -> None:
    next_id = 0
    data_sequences = []
    if len(file_paths) == 0:
        raise ValueError("Specify at least one file")

    for file_path in file_paths:
        with open(file_path, "r") as f:
            str_cells = read_file(f, file_path.split(".")[-1], sep)
        data_sequences.extend(parse(str_cells, next_id))
        next_id = len(data_sequences)
    _plot(data_sequences, x_label, y_label, x_scale, y_scale, x_lim, y_lim, legend_loc)


def plot_inline(
        data_sequences: list[SimpleDataSequence],
        x_label: str | None = None,
        y_label: str | None = None,
        x_scale: str = "linear",
        y_scale: str = "linear",
        x_lim: tuple[float, float] | None = None,
        y_lim: tuple[float, float] | None = None,
        legend_loc: str = "lower"):
    identified_data_sequences = [data_sequences[i].to_data_sequence(i) for i in range(len(data_sequences))]
    _plot(identified_data_sequences, x_label, y_label, x_scale, y_scale, x_lim, y_lim, legend_loc)


def _plot(
        data_sequences: list[DataSequence],
        x_label: str | None = None,
        y_label: str | None = None,
        x_scale: str = "linear",
        y_scale: str = "linear",
        x_lim: tuple[float, float] | None = None,
        y_lim: tuple[float, float] | None = None,
        legend_loc: str = "lower"):
    if x_label is None:
        x_labels = set(seq.x_name for seq in data_sequences)
        if len(x_labels) == 1:
            x_label = x_labels.pop()
    if y_label is None:
        y_labels = set(seq.name for seq in data_sequences)
        if len(y_labels) == 1:
            y_label = y_labels.pop()

    x_lim = (None, None) if x_lim is None else x_lim
    x_axis = DataAxis(DataScaleType(x_scale), x_label, x_lim[0], x_lim[1])
    y_lim = (None, None) if y_lim is None else y_lim
    y_axis = DataAxis(DataScaleType(y_scale), y_label, y_lim[0], y_lim[1])

    data = Data(data_sequences, x_axis, y_axis, DataLegendLoc(legend_loc))
    canvas = data.to_canvas(Canvas)
    terminal_ = canvas.to_terminal(Terminal)
    terminal_.plot()


def main():
    parser = argparse.ArgumentParser(
        prog="scatterminal",
        description="Plot scatter plot on terminal",
    )
    parser.add_argument(
        "file_path",
        nargs="*",
        help="File path to plot (multiple designations possible)"
    )
    parser.add_argument(
        "--sep",
        help="Separator character. If not specified, it is inferred from the extension."
    )
    parser.add_argument(
        "--xscale",
        choices=[scale.value for scale in DataScaleType],
        help="Scale type of x axis",
        default="linear"
    )
    parser.add_argument(
        "--yscale",
        choices=[scale.value for scale in DataScaleType],
        help="Scale type of y axis",
        default="linear"
    )
    parser.add_argument(
        "--xlabel",
        help="Label of x axis"
    )
    parser.add_argument(
        "--ylabel",
        help="Label of y axis"
    )
    parser.add_argument(
        "--xlim",
        nargs=2,
        type=float,
        metavar=("X_MIN", "X_MAX"),
        help="Range of x axis (specify both X_MIN and X_MAX)"
    )
    parser.add_argument(
        "--ylim",
        nargs=2,
        type=float,
        metavar=("Y_MIN", "Y_MAX"),
        help="Range of axis (specify both Y_MIN and Y_MAX)"
    )
    parser.add_argument(
        "--legend-loc",
        choices=[loc.value for loc in DataLegendLoc],
        help="Position of legend",
        default="lower"
    )

    argv = parser.parse_args()
    plot_csv(
        file_paths=argv.file_path,
        sep=argv.sep,
        x_label=argv.xlabel,
        y_label=argv.ylabel,
        x_scale=argv.xscale,
        y_scale=argv.yscale,
        x_lim=argv.xlim,
        y_lim=argv.ylim,
        legend_loc=argv.legend_loc
    )



from io import StringIO

import pytest

from scatterminal.csv_parser import read_file, parse
from scatterminal.data_layer_model import DataSequence


def test_read_file_csv():
    file_obj = StringIO(
        "x,y,z\n"
        "0,1,3\n"
        "2,1,8\n"
        "4,2,13\n"
    )
    expected = [
        ["x", "y", "z"],
        ["0", "1", "3"],
        ["2", "1", "8"],
        ["4", "2", "13"],
    ]
    actual = read_file(file_obj, "csv", None)
    assert expected == actual


def test_read_file_csv_specified_sep():
    file_obj = StringIO(
        "x,y,z\n"
        "0,1,3\n"
        "2,1,8\n"
        "4,2,13\n"
    )
    expected = [
        ["x", "y", "z"],
        ["0", "1", "3"],
        ["2", "1", "8"],
        ["4", "2", "13"],
    ]
    actual = read_file(file_obj, "txt", ",")
    assert expected == actual


def test_read_file_tsv():
    file_obj = StringIO(
        "x\ty\tz\n"
        "0\t1\t3\n"
        "2\t1\t8\n"
        "4\t2\t13\n"
    )
    expected = [
        ["x", "y", "z"],
        ["0", "1", "3"],
        ["2", "1", "8"],
        ["4", "2", "13"],
    ]
    actual = read_file(file_obj, "tsv", None)
    assert expected == actual


def test_read_file_tsv_specified_sep():
    file_obj = StringIO(
        "x\ty\tz\n"
        "0\t1\t3\n"
        "2\t1\t8\n"
        "4\t2\t13\n"
    )
    expected = [
        ["x", "y", "z"],
        ["0", "1", "3"],
        ["2", "1", "8"],
        ["4", "2", "13"],
    ]
    actual = read_file(file_obj, "tsv", "\t")
    assert expected == actual


def test_read_file_csv_file_type_error():
    file_obj = StringIO(
        "x,y,z\n"
        "0,1,3\n"
        "2,1,8\n"
        "4,2,13\n"
    )
    with pytest.raises(ValueError) as e:
        _ = read_file(file_obj, "txt", None)
    assert str(e.value) == "Failed to estimate separator character. Please specify sep explicitly."


@pytest.mark.parametrize(
    ("str_cells", "next_id", "expected"),
    [
        (
                # single column, no label
                [["42.1"], ["42.2"]],
                0,
                [DataSequence([0, 1], [42.1, 42.2], 0, None)]
        ),
        (
                # single column, label specified
                [["value"], ["42.1"], ["42.2"]],
                0,
                [DataSequence([0, 1], [42.1, 42.2], 0, "value")]
        ),
        (
                # double column, no label
                [["0.5", "42.1"], ["1.5", "42.2"]],
                0,
                [DataSequence([0.5, 1.5], [42.1, 42.2], 0, None, None)]
        ),
        (
                # double column, label specified
                [["x", "y"], ["0.5", "42.1"], ["1.5", "42.2"]],
                0,
                [DataSequence([0.5, 1.5], [42.1, 42.2], 0, "y", "x")]
        ),
        (
                # triple column, no label
                [["0.5", "42.1", "100.4"], ["1.5", "42.2", "103.5"]],
                0,
                [DataSequence([0.5, 1.5], [42.1, 42.2], 0, None, None), DataSequence([0.5, 1.5], [100.4, 103.5], 1, None, None)]
        ),
        (
                # triple column, label specified
                [["x", "y1", "y2"], ["0.5", "42.1", "100.4"], ["1.5", "42.2", "103.5"]],
                0,
                [DataSequence([0.5, 1.5], [42.1, 42.2], 0, "y1", "x"), DataSequence([0.5, 1.5], [100.4, 103.5], 1, "y2", "x")]
        ),
        (
                # triple column, label specified, next_id shifted
                [["x", "y1", "y2"], ["0.5", "42.1", "100.4"], ["1.5", "42.2", "103.5"]],
                5,
                [DataSequence([0.5, 1.5], [42.1, 42.2], 5, "y1", "x"), DataSequence([0.5, 1.5], [100.4, 103.5], 6, "y2", "x")]
        ),
    ]
)
def test_parse(str_cells: list[list[str]], next_id: int, expected: list[DataSequence]):
    actual = parse(str_cells, next_id)
    assert actual == expected


def test_parse_not_aligned_error():
    str_cells = [
        ["x", "y"],
        ["0", "1"],
        ["1"],
    ]
    with pytest.raises(ValueError) as e:
        _ = parse(str_cells, 0)
    assert str(e.value) == "The length of column is not aligned."


def test_parse_type_error():
    str_cells = [
        ["x", "y"],
        ["0", "1"],
        ["1", "A"],
    ]
    with pytest.raises(TypeError) as e:
        _ = parse(str_cells, 0)
    assert str(e.value) == "int or float type are only available."



from typing import TextIO

from scatterminal.data_layer_model import DataSequence

ValueType = str | int | float


def read_file(file_obj: TextIO, ext: str, sep: str | None) -> list[list[str]]:
    if sep is None:
        if ext == "csv":
            sep = ","
        elif ext == "tsv":
            sep = "\t"
        else:
            raise ValueError("Failed to estimate separator character. Please specify sep explicitly.")

    lines = file_obj.readlines()

    cell_lines = []
    for line in lines:
        cells = [cell.strip() for cell in line.split(sep)]
        cell_lines.append(cells)
    return cell_lines


def _parse_cell(v: str) -> ValueType:
    if len(v) == 0:
        return float("nan")
    if v.isalpha():
        return v
    if v.isdigit():
        return int(v)
    try:
        return float(v)
    except ValueError:
        return v


def _check_col_num(str_cells: list[list[str]]):
    column_nums = set(len(line) for line in str_cells)
    if len(column_nums) != 1:
        raise ValueError("The length of column is not aligned.")


def _check_col_type(column: list[ValueType]):
    if isinstance(column[0], str):
        value_column = column[1:]
    else:
        value_column = column

    numeric_type_set = {float, int}
    type_set = set(type(cell) for cell in value_column)
    if not type_set.issubset(numeric_type_set):
        raise TypeError("int or float type are only available.")


def _has_header(column: list[ValueType]) -> bool:
    return isinstance(column[0], str)


def parse(str_cells: list[list[str]], next_id: int) -> list[DataSequence]:
    _check_col_num(str_cells)
    # parse and transpose
    parsed_cells = [[_parse_cell(line[i]) for line in str_cells] for i in range(len(str_cells[0]))]

    _ = tuple(_check_col_type(col) for col in parsed_cells)
    col_num = len(parsed_cells)

    has_header = any(_has_header(col) for col in parsed_cells)
    header_line = str_cells[0] if has_header else [None] * col_num
    content_slice = slice(int(has_header), None)

    if col_num == 1:
        content = parsed_cells[0][content_slice]
        return [DataSequence(list(range(len(content))), content, next_id, header_line[0], None)]
    return [
        DataSequence(
            parsed_cells[0][content_slice],
            parsed_cells[i][content_slice],
            next_id + i - 1,
            header_line[i],
            header_line[0]
        ) for i in range(1, col_num)
    ]



from __future__ import annotations

from typing import Callable, Type
import abc
import dataclasses
from enum import Enum
import warnings

from scatterminal.common import log, abs_to_rel
import scatterminal.canvas_layer_model as canvas


class CanvasConvertible(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def to_canvas(self, canvas_type: Type[canvas.TerminalConvertible]) -> canvas.TerminalConvertible:
        pass


@dataclasses.dataclass(frozen=True)
class SimpleDataSequence:
    x: list[int | float]
    y: list[int | float]
    name: str | None = None

    def to_data_sequence(self, seq_id: int) -> DataSequence:
        return DataSequence(self.x, self.y, seq_id, self.name)


@dataclasses.dataclass(frozen=True)
class DataSequence:
    x: list[int | float]
    y: list[int | float]
    seq_id: int
    name: str | None = None
    x_name: str | None = None

    def __post_init__(self):
        if len(self.x) != len(self.y):
            raise ValueError(
                "length of x and y must be equal: (seq_id={0}, len(x)={1}, len(y)={2})".format(self.seq_id, len(self.x), len(self.y))
            )
        if not all(map(lambda x: isinstance(x, (int, float)), self.x)):
            raise TypeError("x must be list[int | float]: seq_id={0}".format(self.seq_id))
        if not all(map(lambda y: isinstance(y, (int, float)), self.y)):
            raise TypeError("y must be list[int | float]: seq_id={0}".format(self.seq_id))
        if (self.name is not None) and (not self.name.isascii()):
            raise ValueError("Sequence name must be ascii: seq_id={0}".format(self.seq_id))
        if (self.x_name is not None) and (not self.x_name.isascii()):
            raise ValueError("Sequence x_name must be ascii: seq_id={0}".format(self.seq_id))

    def create_filtered(self, filter_func: Callable[[tuple[int | float, int | float]], bool]) -> DataSequence:
        new_x = []
        new_y = []
        for xy_pair in filter(filter_func, zip(self.x, self.y)):
            new_x.append(xy_pair[0])
            new_y.append(xy_pair[1])
        return DataSequence(new_x, new_y, self.seq_id, self.name)


class DataScaleType(str, Enum):
    linear = "linear"
    log = "log"


@dataclasses.dataclass(frozen=True)
class DataAxis:
    scale: DataScaleType = DataScaleType.linear
    name: str | None = None
    min_: int | float | None = None
    max_: int | float | None = None

    def __post_init__(self):
        if (self.name is not None) and (not self.name.isascii()):
            raise ValueError("Axis name must be ascii: %s" % self.name)
        if (self.min_ is None) != (self.max_ is None):
            raise ValueError("Axis min and Axis max should be defined simultaneously: (min, max)=(%s, %s)" % (self.min_, self.max_))
        if (self.min_ is not None) and (self.max_ is not None) and self.min_ >= self.max_:
            raise ValueError("Axis min is larger than max: (min, max)=(%f, %f)" % (self.min_, self.max_))
        if self.scale == DataScaleType.log:
            if (self.min_ is not None) and (self.min_ <= 0):
                raise ValueError("Axis min must be positive value on log-scale")


class DataLegendLoc(str, Enum):
    none = "none"
    lower = "lower"
    right = "right"


@dataclasses.dataclass(frozen=True)
class Data(CanvasConvertible):
    data: list[DataSequence]
    x_axis: DataAxis
    y_axis: DataAxis
    legend_loc: DataLegendLoc

    def __post_init__(self):
        if self.x_axis.scale == DataScaleType.log:
            for datum in self.data:
                if min(datum.x) <= 0:
                    warnings.warn(
                        "Non-positive value is detected on log-scale x axis. This data point is not plotted.: seq_id=%d" % datum.seq_id,
                        UserWarning
                    )
        if self.y_axis.scale == DataScaleType.log:
            for datum in self.data:
                if min(datum.y) <= 0:
                    warnings.warn(
                        "Non-positive value is detected on log-scale y axis. This data point is not plotted.: seq_id=%d" % datum.seq_id,
                        UserWarning
                    )

    def to_canvas(self, canvas_type: Type[canvas.TerminalConvertible]) -> canvas.TerminalConvertible:
        # positive-pass filter
        if self.x_axis.scale == DataScaleType.log:
            filtered_data = [datum.create_filtered(lambda xy: xy[0] > 0) for datum in self.data]
        else:
            filtered_data = self.data

        if self.y_axis.scale == DataScaleType.log:
            filtered_data = [datum.create_filtered(lambda xy: xy[1] > 0) for datum in filtered_data]
        else:
            filtered_data = filtered_data

        is_x_range_undef = self.x_axis.min_ is None
        if is_x_range_undef:
            x_min = min(min(datum.x) for datum in filtered_data)
            x_max = max(max(datum.x) for datum in filtered_data)
        else:
            x_min = self.x_axis.min_
            x_max = self.x_axis.max_
        
        edge_space_ratio = 0.1

        if is_x_range_undef:
            if self.x_axis.scale == DataScaleType.linear:
                white_delta = (x_max - x_min) * edge_space_ratio
                canvas_x_range = (x_min - white_delta, x_max + white_delta)
            else:
                white_delta_ratio = log(x_max/x_min) * edge_space_ratio
                canvas_x_range = (x_min / white_delta_ratio, x_max * white_delta_ratio)
        else:
            canvas_x_range = (self.x_axis.min_, self.x_axis.max_)

        is_y_range_undef = self.y_axis.min_ is None
        if is_y_range_undef:
            y_min = min(min(datum.y) for datum in filtered_data)
            y_max = max(max(datum.y) for datum in filtered_data)
        else:
            y_min = self.y_axis.min_
            y_max = self.y_axis.max_

        if is_y_range_undef:
            if self.y_axis.scale == DataScaleType.linear:
                white_delta = (y_max - y_min) * edge_space_ratio
                canvas_y_range = (y_min - white_delta, y_max + white_delta)
            else:
                white_delta_ratio = (log(y_max/y_min) * edge_space_ratio) + 1
                canvas_y_range = (y_min / white_delta_ratio, y_max * white_delta_ratio)
        else:
            canvas_y_range = (self.y_axis.min_, self.y_axis.max_)

        canvas_markers = []
        canvas_legend_elements = []
        x_process_func = log if self.x_axis.scale == DataScaleType.log else lambda v: v
        y_process_func = log if self.y_axis.scale == DataScaleType.log else lambda v: v

        for seq in filtered_data:
            # marker 追加
            for x, y in zip(seq.x, seq.y):
                canvas_markers.append(
                    canvas.CanvasMarker(
                        abs_to_rel(
                            x_process_func(x),
                            x_process_func(canvas_x_range[1]) - x_process_func(canvas_x_range[0]),
                            x_process_func(canvas_x_range[0])
                        ),
                        abs_to_rel(
                            y_process_func(y),
                            y_process_func(canvas_y_range[1]) - y_process_func(canvas_y_range[0]),
                            y_process_func(canvas_y_range[0])
                        ),
                        seq.seq_id
                    )
                )

            # legend 追加
            canvas_legend_elements.append(
                canvas.CanvasLegendElement(seq.seq_id, seq.name)
            )

        canvas_legend = canvas.CanvasLegend(canvas_legend_elements, canvas.CanvasLegendLoc(self.legend_loc))

        # NOTE: データの min, max をそのまま渡している
        canvas_x_axis = canvas.CanvasAxis(
            canvas_x_range[0],
            canvas_x_range[1],
            canvas.CanvasScaleType.linear if self.x_axis.scale == DataScaleType.linear else canvas.CanvasScaleType.log,
            self.x_axis.name
        )
        canvas_y_axis = canvas.CanvasAxis(
            canvas_y_range[0],
            canvas_y_range[1],
            canvas.CanvasScaleType.linear if self.y_axis.scale == DataScaleType.linear else canvas.CanvasScaleType.log,
            self.y_axis.name
        )
        return canvas.Canvas(
            canvas_markers,
            canvas_x_axis,
            canvas_y_axis,
            canvas_legend
        )



from typing import Type
import abc
import dataclasses
from enum import Enum
import shutil
import os
import math
import warnings

from scatterminal.common import log, exp, abs_to_rel
import scatterminal.terminal_layer_model as terminal


def _quantize(rel_value: float, grid_num: int) -> int:
    return round(rel_value * (grid_num - 1))


class TerminalConvertible(metaclass=abc.ABCMeta):
    @abc.abstractmethod
    def to_terminal(self, plot_type: Type[terminal.Plottable]) -> terminal.Plottable:
        pass


@dataclasses.dataclass(frozen=True)
class CanvasPoint:
    x: float
    y: float


@dataclasses.dataclass(frozen=True)
class CanvasMarker(CanvasPoint):
    marker_group_id: int


class CanvasScaleType(str, Enum):
    linear = "linear"
    log = "log"


class _TerminalSize:
    def __init__(
            self,
            terminal_size: os.terminal_size,
            y_tick_label_size: int,  # x offset 1
            has_y_axis_label: bool,  # y offset 1
            right_legend_size: int,  # x offset 2
            lower_legend_size: int,  # y offset 2
    ):
        self.terminal_size = terminal_size
        self._has_y_axis_label = has_y_axis_label

        self._tot_y_upper_offset = 1
        self._tot_y_lower_offset = lower_legend_size + 3  # axis_label, empty, tick_label
        self._tot_x_left_offset = y_tick_label_size + int(has_y_axis_label) + 1  # empty
        self._tot_x_right_offset = right_legend_size

    @property
    def lines(self) -> int:
        return self.terminal_size.lines

    @property
    def columns(self) -> int:
        return self.terminal_size.columns

    @property
    def canvas_lines(self) -> int:
        return self.terminal_size.lines - self._tot_y_lower_offset - self._tot_y_upper_offset

    @property
    def canvas_columns(self) -> int:
        return self.terminal_size.columns - self._tot_x_left_offset - self._tot_x_right_offset

    @property
    def has_y_axis_label(self) -> bool:
        return self._has_y_axis_label

    def from_canvas_to_terminal_columns(self, val: int) -> int:
        return val + self._tot_x_left_offset

    def from_canvas_to_terminal_lines(self, val: int) -> int:
        return val + self._tot_y_lower_offset


@dataclasses.dataclass(frozen=True)
class CanvasAxis:
    min_: float
    max_: float
    scale: CanvasScaleType = CanvasScaleType.linear
    name: str | None = None

    def gen_y_axis(
            self,
            label_and_values: list[tuple[str, float]],
            terminal_size: _TerminalSize
    ) -> terminal.TerminalYAxis:
        min_and_max = self._gen_axis_min_max(label_and_values)
        axis_label_offset = int(terminal_size.has_y_axis_label) * 2

        tick_labels = []
        tick_grid_set = set()
        for label, val in label_and_values:
            if self.scale == CanvasScaleType.linear:
                rel = abs_to_rel(val, min_and_max[1] - min_and_max[0], min_and_max[0])
            else:
                rel = abs_to_rel(log(val), log(min_and_max[1]) - log(min_and_max[0]), log(min_and_max[0]))
            quantized = _quantize(rel, terminal_size.canvas_lines)
            terminal_quantized = terminal_size.from_canvas_to_terminal_lines(quantized)
            tick_labels.append(
                terminal.TerminalLabel(axis_label_offset, terminal_quantized, label)
            )
            tick_grid_set.add(quantized)

        axis_lines = []
        for i in range(terminal_size.canvas_lines):
            if i in tick_grid_set:
                char = "+"
            else:
                char = "|"
            axis_lines.append(
                terminal.TerminalMarker(
                    terminal_size.from_canvas_to_terminal_columns(0),
                    terminal_size.from_canvas_to_terminal_lines(i),
                    char
                )
            )

        if self.name is None:
            axis_label = None
        else:
            if len(self.name) > terminal_size.canvas_lines:
                raise ValueError("Too large y axis label: size=%d, max=%d" % (len(self.name), terminal_size.canvas_lines))

            start_y = (terminal_size.lines + len(self.name)) // 2
            axis_label = []
            for i in range(len(self.name)):
                axis_label.append(
                    terminal.TerminalMarker(0, start_y - i, self.name[i])
                )
        return terminal.TerminalYAxis(axis_lines, tick_labels, axis_label)

    def gen_x_axis(
            self,
            label_and_values: list[tuple[str, float]],
            terminal_size: _TerminalSize
    ) -> terminal.TerminalXAxis:
        min_and_max = self._gen_axis_min_max(label_and_values)

        tick_labels = []
        tick_grid_set = set()
        for label, val in label_and_values:
            if self.scale == CanvasScaleType.linear:
                rel = abs_to_rel(val, min_and_max[1] - min_and_max[0], min_and_max[0])
            else:
                rel = abs_to_rel(log(val), log(min_and_max[1]) - log(min_and_max[0]), log(min_and_max[0]))
            quantized = _quantize(rel, terminal_size.canvas_columns)
            terminal_quantized_shifted = terminal_size.from_canvas_to_terminal_columns(quantized) - len(label) // 2
            tick_labels.append(
                terminal.TerminalLabel(
                    terminal_quantized_shifted,
                    terminal_size.from_canvas_to_terminal_lines(-1),
                    label,
                    allow_left_shift=True
                )
            )
            tick_grid_set.add(quantized)

        axis_lines = []
        for i in range(terminal_size.canvas_columns):
            if i in tick_grid_set:
                char = "+"
            else:
                char = "-"
            axis_lines.append(
                terminal.TerminalMarker(
                    terminal_size.from_canvas_to_terminal_columns(i),
                    terminal_size.from_canvas_to_terminal_lines(0),
                    char
                )
            )

        if self.name is None:
            axis_label = None
        else:
            if len(self.name) > terminal_size.canvas_columns:
                raise ValueError("Too large x axis label: size=%d, max=%d" % (len(self.name), terminal_size.canvas_columns))
            canvas_center_x = terminal_size.from_canvas_to_terminal_columns(terminal_size.canvas_columns // 2) - len(self.name) // 2
            axis_label = terminal.TerminalLabel(canvas_center_x, terminal_size.from_canvas_to_terminal_lines(-2), self.name)
        return terminal.TerminalXAxis(axis_lines, tick_labels, axis_label)

    def calc_tick(self, primary_limit: int = 20, secondary_limit: int = 10) -> list[tuple[str, float]]:
        min_and_max = (self.min_, self.max_)
        if self.scale == CanvasScaleType.linear:
            ticks = self._calc_linear_tick(*min_and_max, *(primary_limit, secondary_limit))
        else:
            ticks = self._calc_log_tick(*min_and_max, *(primary_limit, secondary_limit))

        use_index = self._use_index(*min_and_max)
        sd = self._calc_significant_digits(*min_and_max)
        format_ = "e" if use_index else "f"
        return [(f"{tick:.{sd}{format_}}", tick) for tick in ticks]

    @staticmethod
    def _calc_linear_tick(min_: float, max_: float, primary_limit: int, secondary_limit: int) -> list[float]:
        value_range = max_ - min_
        tick_scale = exp(round(log(value_range) - 1))
        tick_num = value_range / tick_scale
        if tick_num > primary_limit:
            tick_scale *= 5
            tick_num /= 5
        elif tick_num > secondary_limit:
            tick_scale *= 2.5
            tick_num /= 2.5
        tick_num = int(math.ceil(tick_num))

        min_tick = round(min_ / tick_scale) * tick_scale
        return [min_tick + (i * tick_scale) for i in range(tick_num)]

    @staticmethod
    def _calc_log_tick(min_: float, max_: float, primary_limit: int, secondary_limit: int) -> list[float]:
        value_range = log(max_) - log(min_)
        tick_scale = exp(round(log(value_range) - 1))
        tick_num = value_range / tick_scale
        if tick_num > primary_limit:
            tick_scale *= 5
            tick_num /= 5
        elif tick_num > secondary_limit:
            tick_scale *= 2.5
            tick_num /= 2.5
        tick_num = int(math.ceil(tick_num))

        min_tick = round(log(min_) / tick_scale) * tick_scale
        return [exp(min_tick + (i * tick_scale)) for i in range(tick_num)]

    @staticmethod
    def _calc_significant_digits(min_: float, max_: float) -> int:
        delta = max_ - min_
        inv_med = 2/(abs(min_) + abs(max_))

        raw_digit = log(delta * inv_med) + 2
        if raw_digit < 1:
            return 1
        if raw_digit > 6:
            return 6
        return int(math.ceil(raw_digit))

    def _use_index(self, min_: float, max_: float) -> bool:
        far_from_origin = abs(min_) if abs(min_) > abs(max_) else abs(max_)
        nea_from_origin = abs(min_) if abs(min_) < abs(max_) else abs(max_)

        if self.scale == CanvasScaleType.linear:
            median_order = log((far_from_origin + nea_from_origin) / 2)
        else:
            median_order = log(math.sqrt(far_from_origin * nea_from_origin))

        return median_order > 4 or median_order < -4

    def _gen_axis_min_max(self, label_and_values: list[tuple[str, float]]) -> tuple[float, float]:
        return (
            min(min(lv[1] for lv in label_and_values), self.min_),
            max(max(lv[1] for lv in label_and_values), self.max_),
        )


@dataclasses.dataclass(frozen=True)
class CanvasLegendElement:
    marker_group_id: int
    sequence_name: str | None


class CanvasLegendLoc(str, Enum):
    none = "none"
    lower = "lower"
    right = "right"


@dataclasses.dataclass(frozen=True)
class CanvasLegend:
    legend_elements: list[CanvasLegendElement]
    loc: CanvasLegendLoc

    def gen_right_legend(self, marker_char_dict: dict[int, str]) -> tuple[tuple[int, int], terminal.TerminalLegend]:
        legend_element_strings = [f"{marker_char_dict[le.marker_group_id]}: {le.sequence_name}" for le in self.legend_elements]
        max_legend_size = max(map(len, legend_element_strings))

        _terminal_size = shutil.get_terminal_size()
        legend_labels = []
        for i_line in range(len(legend_element_strings)):
            x = _terminal_size.columns - max_legend_size
            y = _terminal_size.lines - i_line - 2
            legend_labels.append(
                terminal.TerminalLabel(x, y, legend_element_strings[i_line])
            )

        return (max_legend_size, 0), terminal.TerminalLegend(legend_labels)

    def gen_lower_legend(self, marker_char_dict: dict[int, str]) -> tuple[tuple[int, int], terminal.TerminalLegend]:
        legend_element_strings = [f"{marker_char_dict[le.marker_group_id]}: {le.sequence_name}" for le in self.legend_elements]
        max_legend_size = max(map(len, legend_element_strings))

        space = 2
        _terminal_size = shutil.get_terminal_size()
        max_legend_num_per_line = _terminal_size.columns // (max_legend_size + space)  # 1行に表示できる最大legend数
        legend_line_num = int(math.ceil(len(legend_element_strings) / max_legend_num_per_line))
        legend_labels = []
        for i_line in range(legend_line_num):
            y = legend_line_num - i_line - 1
            x = 0
            for j_legend in range(max_legend_num_per_line):
                index = max_legend_num_per_line * i_line + j_legend
                if index >= len(legend_element_strings):
                    # end line
                    break
                legend = legend_element_strings[index]
                legend_labels.append(
                    terminal.TerminalLabel(x, y, legend)
                )
                x += len(legend) + space

        return (0, legend_line_num), terminal.TerminalLegend(legend_labels)


@dataclasses.dataclass(frozen=True)
class Canvas(TerminalConvertible):
    markers: list[CanvasMarker]
    x_axis: CanvasAxis
    y_axis: CanvasAxis
    legend: CanvasLegend

    def to_terminal(self, plot_type: Type[terminal.Plottable]) -> terminal.Plottable:
        # generate marker dict
        marker_char_dict = self._gen_marker_char_dict(set(marker.marker_group_id for marker in self.markers), plot_type.get_marker_chars())

        # generate legend
        if self.legend.loc == CanvasLegendLoc.right:
            legend_offsets, terminal_legend = self.legend.gen_right_legend(marker_char_dict)
        elif self.legend.loc == CanvasLegendLoc.lower:
            legend_offsets, terminal_legend = self.legend.gen_lower_legend(marker_char_dict)
        else:
            legend_offsets: tuple[int, int] = 0, 0
            terminal_legend = None

        # generate y axis
        tick_label_and_values_y = self.y_axis.calc_tick()
        max_label_size_y = max(len(tick_label_and_value[0]) for tick_label_and_value in tick_label_and_values_y)
        terminal_size = _TerminalSize(shutil.get_terminal_size(), max_label_size_y, self.y_axis.name is not None, *legend_offsets)
        terminal_y_axis = self.y_axis.gen_y_axis(tick_label_and_values_y, terminal_size)

        # generate x axis
        primary, secondary = 18, 8
        tick_label_and_values_x = self.x_axis.calc_tick()
        total_label_size_x = sum(len(tick_label_and_value[0]) for tick_label_and_value in tick_label_and_values_x)
        while total_label_size_x > (terminal_size.columns / 2):
            primary -= 1
            secondary -= 1
            tick_label_and_values_x = self.x_axis.calc_tick(primary, secondary)
            total_label_size_x = sum(len(tick_label_and_value[0]) for tick_label_and_value in tick_label_and_values_x)
        terminal_x_axis = self.x_axis.gen_x_axis(tick_label_and_values_x, terminal_size)

        # generate marker
        terminal_markers = []
        for marker in self.markers:
            x = _quantize(marker.x, terminal_size.canvas_columns)
            y = _quantize(marker.y, terminal_size.canvas_lines)
            terminal_mark = terminal.TerminalMarker(
                x=terminal_size.from_canvas_to_terminal_columns(x),
                y=terminal_size.from_canvas_to_terminal_lines(y),
                char=marker_char_dict[marker.marker_group_id]
            )
            terminal_markers.append(terminal_mark)

        return terminal.Terminal(
            terminal_size.lines, terminal_size.columns,
            terminal_markers, terminal_x_axis, terminal_y_axis, terminal_legend
        )

    @staticmethod
    def _gen_marker_char_dict(marker_group_ids: set[int], chars: list[str]) -> dict[int, str]:
        char_num = len(chars)
        if len(marker_group_ids) > char_num:
            warnings.warn("The number of data series exceeds the number of marker types available. "
                          "Then, Markers are reused making data identification difficult.", UserWarning)

        char_dict = {}
        for gid in marker_group_ids:
            char_dict[gid] = chars[gid % char_num]
        return char_dict
